<?php

namespace Kash;

use Kash\Serialize\ObjectSerialize;
use Laravel\SerializableClosure\Contracts\Serializable;
use Psr\SimpleCache\CacheException;
use Psr\SimpleCache\CacheInterface;
use Psr\SimpleCache\InvalidArgumentException;

use Ulmus\EntityCollection;
use function apcu_fetch, apcu_store, apcu_delete, apcu_clear_cache, apcu_exists, apcu_enabled;

class ApcuCache implements CacheInterface
{
    const DEFAULT_NS_SEPARATOR = ':';

    public function __construct(
        protected CacheInvalidator $invalidator,
        protected string $namespace,
        protected int $ttl = 3600,
    ) {
        if (! apcu_enabled() ) {
            throw new class extends \Exception implements CacheException{
                public function __construct(string $message = "APCu extension is required to operate this class. If you operate this script from the CLI, use make sure apc.enable_cli is enabled in your php.ini (or set by your server).", int $code = 0, ? \Throwable $previous = null)
                {
                    parent::__construct($message, $code, $previous);
                }
            };
        }
    }

    public function get(string $key, mixed $default = null) : mixed
    {
        $value = apcu_fetch($this->fullKey($key), $status);

        if ($value instanceof Serialize\ObjectSerialize) {
            $value = $value->restore();
        }

        return $status === true ? $value : $default;
    }

    public function set(string $key, mixed $value, null|int|\DateInterval $ttl = null) : bool
    {
        /*if ( is_object($value) && ! ( $value instanceof Serialize\ObjectSerialize )) {
            $value = $this->serializeObject($value);
        }*/

        return apcu_store($this->fullKey($key), $value, $this->handleTTL($ttl));
    }

    public function delete(string $key) : bool
    {
        return apcu_delete($this->fullKey($key));
    }

    public function clear() : bool
    {
        return apcu_clear_cache();
    }

    public function getMultiple(iterable $keys, mixed $default = null) : iterable
    {
        $arr = iterator_to_array($keys);

        $nsKeys = $this->fullKey($keys);

        $result = apcu_fetch($nsKeys);

        return array_combine(array_map(fn($e) => substr($e, strlen($this->namespace) + strlen(static::DEFAULT_NS_SEPARATOR)), $nsKeys), $result) + array_fill_keys($arr, $default);
    }

    public function setMultiple(iterable $values, null|int|\DateInterval $ttl = null): bool
    {
        $arr = iterator_to_array($values);

        $list = array_combine(array_map( [$this, 'prependNamespace' ], array_keys($arr)), $arr);

        $result = apcu_store($list, $this->handleTTL($ttl));

        return in_array($result, [ [], true ], true);
    }

    public function deleteMultiple(iterable $keys) : bool
    {
        return apcu_delete($this->fullKey($keys)) === [];
    }

    public function has(string $key) : bool
    {
        return apcu_exists($this->fullKey($key));
    }

    protected function handleTTL(null|int|\DateInterval $ttl) : int
    {
        if ($ttl instanceof \DateInterval) {
            $ttl = (new \DateTime)->add($ttl)->getTimestamp() - (new \DateTime)->getTimestamp();
        }

        $ttl ??= $this->ttl;

        return (int) $ttl;
    }

    protected function fullKey(string|array $key) : string|array
    {
        return is_array($key) ? array_map([ $this, 'prependNamespace' ], $key) : $this->prependNamespace($key);
    }

    protected function prependNamespace(string $key) : string
    {
        return $this->invalidator . $this->namespace . static::DEFAULT_NS_SEPARATOR . $key;
    }

    protected function serializeObject(object $input)
    {
        $objSerialize = new ObjectSerialize();
        return $objSerialize->transform($input);
    }
}
